/*
 * Socket.c 
 *
 * Kevin P. Smith  1/26/89
 *
 * Provides all of the support for sending packets to the client.
 */
#include "copyright2.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <math.h>
#include <errno.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
int handleTorpReq(), handlePhasReq(), handleSpeedReq();
int handleDirReq(), handleShieldReq(), handleRepairReq(), handleOrbitReq();
int handlePractrReq(), handleBombReq(), handleBeamReq(), handleCloakReq();
int handleDetTReq(), handleCopilotReq();
int handleOutfit(), handleLoginReq();
int handlePlasmaReq(), handleWarReq(), handlePlanlockReq();
int handlePlaylockReq(), handleDetMReq();
int handleTractorReq(), handleRepressReq();
int handleCoupReq(), handleRefitReq(), handleMessageReq();
int handleQuitReq(), handleOptionsPacket();
int handleSocketReq(), handleByeReq();
int handleDockingReq(), handleReset();
int handleUpdatesReq(), handleReserved();

static int remoteaddr= -1;		/* inet address in net format */

struct packet_handler handlers[] = {
    { 0, NULL},		/* record 0 */
    { sizeof(struct mesg_cpacket), handleMessageReq },	   /* CP_MESSAGE */
    { sizeof(struct speed_cpacket), handleSpeedReq },	   /* CP_SPEED */
    { sizeof(struct dir_cpacket), handleDirReq },	   /* CP_DIRECTION */
    { sizeof(struct phaser_cpacket), handlePhasReq },	   /* CP_PHASER */
    { sizeof(struct plasma_cpacket), handlePlasmaReq },	   /* CP_PLASMA */
    { sizeof(struct torp_cpacket), handleTorpReq },	   /* CP_TORP */
    { sizeof(struct quit_cpacket), handleQuitReq },	   /* CP_QUIT */
    { sizeof(struct login_cpacket), handleLoginReq },	   /* CP_LOGIN */
    { sizeof(struct outfit_cpacket), handleOutfit },	   /* CP_OUTFIT */
    { sizeof(struct war_cpacket), handleWarReq },	   /* CP_WAR */
    { sizeof(struct practr_cpacket), handlePractrReq },	   /* CP_PRACTR */
    { sizeof(struct shield_cpacket), handleShieldReq },	   /* CP_SHIELD */
    { sizeof(struct repair_cpacket), handleRepairReq },	   /* CP_REPAIR */
    { sizeof(struct orbit_cpacket), handleOrbitReq },	   /* CP_ORBIT */
    { sizeof(struct planlock_cpacket), handlePlanlockReq },/* CP_PLANLOCK */
    { sizeof(struct playlock_cpacket), handlePlaylockReq },/* CP_PLAYLOCK */
    { sizeof(struct bomb_cpacket), handleBombReq },	   /* CP_BOMB */
    { sizeof(struct beam_cpacket), handleBeamReq },	   /* CP_BEAM */
    { sizeof(struct cloak_cpacket), handleCloakReq },	   /* CP_CLOAK */
    { sizeof(struct det_torps_cpacket), handleDetTReq },   /* CP_DET_TORPS */
    { sizeof(struct det_mytorp_cpacket), handleDetMReq },  /* CP_DET_MYTORP */
    { sizeof(struct copilot_cpacket), handleCopilotReq },  /* CP_COPLIOT */
    { sizeof(struct refit_cpacket), handleRefitReq }, 	   /* CP_REFIT */
    { sizeof(struct tractor_cpacket), handleTractorReq },  /* CP_TRACTOR */
    { sizeof(struct repress_cpacket), handleRepressReq },  /* CP_REPRESS */
    { sizeof(struct coup_cpacket), handleCoupReq },	   /* CP_COUP */
    { sizeof(struct socket_cpacket), handleSocketReq },	   /* CP_SOCKET */
    { sizeof(struct options_cpacket), handleOptionsPacket },/* CP_OPTIONS */
    { sizeof(struct bye_cpacket), handleByeReq },	   /* CP_BYE */
    { sizeof(struct dockperm_cpacket), handleDockingReq }, /* CP_DOCKPERM */
    { sizeof(struct updates_cpacket), handleUpdatesReq 	}, /* CP_UPDATES */
    { sizeof(struct resetstats_cpacket), handleReset },    /* CP_RESETSTATS */
    { sizeof(struct reserved_cpacket), handleReserved }	   /* CP_RESERVED */
};

int sizes[] = {
    0,	/* record 0 */
    sizeof(struct mesg_spacket),		/* SP_MESSAGE */
    sizeof(struct plyr_info_spacket), 		/* SP_PLAYER_INFO */
    sizeof(struct kills_spacket),		/* SP_KILLS */
    sizeof(struct player_spacket),		/* SP_PLAYER */
    sizeof(struct torp_info_spacket),		/* SP_TORP_INFO */
    sizeof(struct torp_spacket), 		/* SP_TORP */
    sizeof(struct phaser_spacket),		/* SP_PHASER */
    sizeof(struct plasma_info_spacket),		/* SP_PLASMA_INFO */
    sizeof(struct plasma_spacket),		/* SP_PLASMA */
    sizeof(struct warning_spacket),		/* SP_WARNING */
    sizeof(struct motd_spacket),		/* SP_MOTD */
    sizeof(struct you_spacket),			/* SP_YOU */
    sizeof(struct queue_spacket),		/* SP_QUEUE */
    sizeof(struct status_spacket),		/* SP_STATUS */
    sizeof(struct planet_spacket),		/* SP_PLANET */
    sizeof(struct pickok_spacket),		/* SP_PICKOK */
    sizeof(struct login_spacket),		/* SP_LOGIN */
    sizeof(struct flags_spacket),		/* SP_FLAGS */
    sizeof(struct mask_spacket),		/* SP_MASK */
    sizeof(struct pstatus_spacket),		/* SP_PSTATUS */
    sizeof(struct badversion_spacket),		/* SP_BADVERSION */
    sizeof(struct hostile_spacket),		/* SP_HOSTILE */
    sizeof(struct stats_spacket),		/* SP_STATS */
    sizeof(struct plyr_login_spacket),		/* SP_PL_LOGIN */
    sizeof(struct reserved_spacket), 		/* SP_RESERVED */
    sizeof(struct planet_loc_spacket)		/* SP_PLANET_LOC */
};

#define NUM_PACKETS (sizeof(handlers) / sizeof(handlers[0]) - 1)
#define NUM_SIZES (sizeof(sizes) / sizeof(sizes[0]) - 1)

int packetsReceived[NUM_PACKETS+1] = { 0 };
int packetsSent[NUM_SIZES+1] = { 0 };

#define BUFSIZE 16738
static char buf[BUFSIZE];		/* Socket buffer */
static char *bufptr=buf;
int clientDead=0;

struct plyr_info_spacket clientPlayersInfo[MAXPLAYER];
struct plyr_login_spacket clientLogin[MAXPLAYER];
struct hostile_spacket clientHostile[MAXPLAYER];
struct stats_spacket clientStats[MAXPLAYER];
struct player_spacket clientPlayers[MAXPLAYER];
struct kills_spacket clientKills[MAXPLAYER];
struct flags_spacket clientFlags[MAXPLAYER];
struct pstatus_spacket clientPStatus[MAXPLAYER];
int msgCurrent;
struct torp_info_spacket clientTorpsInfo[MAXPLAYER*MAXTORP];
struct torp_spacket clientTorps[MAXPLAYER*MAXTORP];
struct phaser_spacket clientPhasers[MAXPLAYER];
struct you_spacket clientSelf;
struct status_spacket clientStatus;
struct planet_spacket clientPlanets[MAXPLANETS];
struct planet_loc_spacket clientPlanetLocs[MAXPLANETS];
struct plasma_info_spacket clientPlasmasInfo[MAXPLAYER*MAXPLASMA];
struct plasma_spacket clientPlasmas[MAXPLAYER*MAXPLASMA];
int mustUpdate[MAXPLAYER];

connectToClient(machine, port)
char *machine;
int port;
{
    int ns;
    struct sockaddr_in addr;
    struct hostent *hp;
    int len,cc;
    char buf[BUFSIZ];
    int i;

    if (sock!=-1) {
	shutdown(sock,2);
	sock= -1;
    }
    printf("Connecting to %s through %d\n", machine, port);

    if ((ns=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	printf("I cannot create a socket\n");
	exit(2);
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);

    if (remoteaddr!=-1) {
	addr.sin_addr.s_addr=remoteaddr;
    } else if ((addr.sin_addr.s_addr = inet_addr(machine)) == -1) {
	if ((hp = gethostbyname(machine)) == NULL) {
	    printf("I cannot get host name\n");
	    close(ns);
	    exit(1);
	} else {
	    addr.sin_addr.s_addr = *(long *) hp->h_addr;
	}
    }
    remoteaddr=addr.sin_addr.s_addr;

    if (connect(ns, &addr, sizeof(addr)) < 0) {
	printf("I cannot connect through port %d\n", port);
	close(ns);
	return(0);
    }
    sock=ns;
    initClientData();
    testtime = -1;
    return(1);
}

/* Check the socket to read it's inet addr for possible future use.
 */
checkSocket()
{
    struct sockaddr_in sin;
    int length;

    length=sizeof(sin);
    if (getpeername(sock, (struct sockaddr *) &sin, &length) < 0) {
	/* A bad thing. */
	return;
    }

    remoteaddr=sin.sin_addr.s_addr;
}

initClientData()
/* invalidates all data, so it is all sent to the client */
{
    int i;

    clientDead=0;
    for (i=0; i<MAXPLAYER; i++) {
	clientHostile[i].hostile= -1;
	clientStats[i].losses= -1;
	clientLogin[i].rank= -1;
	clientPlayersInfo[i].shiptype= -1;
	clientPStatus[i].status= -1;
	clientPlayers[i].x= htonl(-1);
	clientPhasers[i].status= -1;
	clientKills[i].kills= htonl(-1);
	clientFlags[i].flags= htonl(-1);
	mustUpdate[i]=0;
    }
    for (i=0; i<MAXPLAYER*MAXTORP; i++) {
	clientTorpsInfo[i].status= -1;
	clientTorps[i].x= -1;
    }
    for (i=0; i<MAXPLAYER*MAXPLASMA; i++) {
	clientPlasmasInfo[i].status= -1;
	clientPlasmas[i].x= -1;
    }
    for (i=0; i<MAXPLANETS; i++) {
	clientPlanets[i].armies= htonl(-1);
	clientPlanetLocs[i].x= htonl(-1);
    }
    msgCurrent=(mctl->mc_current+1) % MAXMESSAGE;
    clientSelf.pnum= -1;
}

isClientDead()
{
    return(clientDead);
}

updateClient()
{
    int i;

    for (i=0; i<MAXPLAYER; i++) {
	mustUpdate[i]=0;
    }
    updateTorps();
    updatePlasmas();
    updateStatus();
    updateSelf();
    updatePhasers();
    updateShips();
    updatePlanets();
    updateMessages();
    if (bufptr==buf) {	
	/* We sent nothing!  We better send something to wake him */
	sendClientPacket(&clientSelf);
    }
    flushSockBuf();
    repCount++;
}

updateStatus()
{
    /* Update status every 10 seconds? */
    if (repCount % efticks(50) == 0 && ntohl(clientStatus.timeprod) != status->timeprod) {
	clientStatus.type=SP_STATUS;
	clientStatus.tourn=status->tourn;
	clientStatus.armsbomb=htonl(status->armsbomb);
	clientStatus.planets=htonl(status->planets);
	clientStatus.kills=htonl(status->kills);
	clientStatus.losses=htonl(status->losses);
	clientStatus.time=htonl(status->time);
	clientStatus.timeprod=htonl(status->timeprod);
	sendClientPacket(&clientStatus);
    }
}

updateSelf()
{
    if (ntohl(clientSelf.fuel) != me->p_fuel ||
	    ntohl(clientSelf.shield) != me->p_shield ||
	    ntohl(clientSelf.damage) != me->p_damage ||
	    ntohs(clientSelf.etemp) != me->p_etemp ||
	    ntohs(clientSelf.wtemp) != me->p_wtemp ||
	    ntohl(clientSelf.flags) != me->p_flags || 
	    clientSelf.armies != me->p_armies ||
	    clientSelf.swar != me->p_swar ||
	    ntohs(clientSelf.whydead) != me->p_whydead ||
	    ntohs(clientSelf.whodead) != me->p_whodead ||
	    clientSelf.pnum != me->p_no) {
	clientSelf.type=SP_YOU;
	clientSelf.pnum=me->p_no;
	clientSelf.flags=htonl(me->p_flags);
	clientSelf.swar=me->p_swar;
	clientSelf.hostile=me->p_hostile;
	clientSelf.armies=me->p_armies;
	clientSelf.shield=htonl(me->p_shield);
	clientSelf.fuel=htonl(me->p_fuel);
	clientSelf.etemp=htons(me->p_etemp);
	clientSelf.wtemp=htons(me->p_wtemp);
	clientSelf.whydead=htons(me->p_whydead);
	clientSelf.whodead=htons(me->p_whodead);
	clientSelf.damage=htonl(me->p_damage);
	sendClientPacket(&clientSelf);
    }
}

updateShips()
{
    register int i;
    register struct player *pl;
    register struct plyr_info_spacket *cpli;
    register struct player_spacket *cpl;
    register struct kills_spacket *kills;
    register struct flags_spacket *flags;
    register struct pstatus_spacket *pstatus;
    register struct plyr_login_spacket *login;
    register struct hostile_spacket *hostile;
    register struct stats_spacket *stats;
    int update;
    int x,y;
#define FLAGMASK (PFSHIELD|PFBOMB|PFORBIT|PFCLOAK|PFROBOT|PFBEAMUP|PFBEAMDOWN|PFPRACTR|PFDOCK|PFTRACT|PFPRESS|PFDOCKOK)

    /* Please excuse the ugliness of this loop declaration */
    for (i=0, pl=players, cpli=clientPlayersInfo, cpl=clientPlayers, kills=clientKills, flags=clientFlags, pstatus=clientPStatus, login=clientLogin, hostile=clientHostile, stats=clientStats;
    	    i<MAXPLAYER; 
	    i++, pl++, cpli++, cpl++, kills++, flags++, pstatus++, login++, hostile++, stats++) {
	update=0;
	if (strcmp(pl->p_name, login->name)!=0 ||
	    pl->p_stats.st_rank != login->rank ||
	    strcmp(pl->p_monitor, login->monitor)!=0) {
	    strncpy(login->name,pl->p_name,15);
	    strncpy(login->monitor,pl->p_monitor,15);
	    strncpy(login->login,pl->p_login,15);
	    login->name[15]=0;
	    login->monitor[15]=0;
	    login->login[15]=0;
	    login->type=SP_PL_LOGIN;
	    login->pnum=i;
	    login->rank=pl->p_stats.st_rank;
	    sendClientPacket(login);
	}
	if ((pl!=me && (pl->p_swar & me->p_team)!=hostile->war) ||
	    (pl!=me && (pl->p_hostile & me->p_team)!=hostile->hostile) ||
	    (pl==me && pl->p_swar != hostile->war) ||
	    (pl==me && pl->p_hostile != hostile->hostile)) {
	    hostile->type=SP_HOSTILE;
	    if (pl==me) {
		hostile->war=pl->p_swar;
		hostile->hostile=pl->p_hostile;
	    } else {
		hostile->war=(pl->p_swar & me->p_team);
		hostile->hostile=(pl->p_hostile & me->p_team);
	    }
	    hostile->pnum=i;
	    sendClientPacket(hostile);
	}
	/* Send stat packets once per five updates.
	 * But, only send one.  We will cycle through them all eventually.
	 */
	if (repCount % (MAXPLAYER * efticks(5)) == i*efticks(5) &&
	    (stats->tkills != htonl(pl->p_stats.st_tkills) ||
	     stats->tlosses != htonl(pl->p_stats.st_tlosses) ||
	     stats->kills != htonl(pl->p_stats.st_kills) ||
	     stats->losses != htonl(pl->p_stats.st_losses) ||
	     stats->tticks != htonl(pl->p_stats.st_tticks) ||
	     stats->tplanets != htonl(pl->p_stats.st_tplanets) ||
	     stats->tarmies != htonl(pl->p_stats.st_tarmsbomb) ||
	     stats->sbkills != htonl(pl->p_stats.st_sbkills) ||
	     stats->sblosses != htonl(pl->p_stats.st_sblosses) ||
	     stats->armies != htonl(pl->p_stats.st_armsbomb) ||
	     stats->planets != htonl(pl->p_stats.st_planets))) {
	    stats->tkills = htonl(pl->p_stats.st_tkills);
	    stats->tlosses = htonl(pl->p_stats.st_tlosses);
	    stats->kills = htonl(pl->p_stats.st_kills);
	    stats->losses = htonl(pl->p_stats.st_losses);
	    stats->tticks = htonl(pl->p_stats.st_tticks);
	    stats->tplanets = htonl(pl->p_stats.st_tplanets);
	    stats->tarmies = htonl(pl->p_stats.st_tarmsbomb);
	    stats->sbkills = htonl(pl->p_stats.st_sbkills);
	    stats->sblosses = htonl(pl->p_stats.st_sblosses);
	    stats->armies = htonl(pl->p_stats.st_armsbomb);
	    stats->planets = htonl(pl->p_stats.st_planets);
	    stats->maxkills=htonl((int) (pl->p_stats.st_maxkills * 100));
	    stats->sbmaxkills=htonl((int) (pl->p_stats.st_sbmaxkills * 100));
	    stats->type=SP_STATS;
	    stats->pnum=i;
	    sendClientPacket(stats);
	}
	if (pl->p_ship.s_type != cpli->shiptype ||
	    pl->p_team != cpli->team) {
	    cpli->type=SP_PLAYER_INFO;
	    cpli->pnum=i;
	    cpli->shiptype=pl->p_ship.s_type;
	    cpli->team=pl->p_team;
	    sendClientPacket(cpli);
	}
	if (kills->kills != htonl((int) (pl->p_kills*100))) {
	    kills->type=SP_KILLS;
	    kills->pnum=i;
	    kills->kills=htonl((int) (pl->p_kills*100));
	    sendClientPacket(kills);
	}
	if (pstatus->status != pl->p_status) {
	    /* We update the location of people whose status has changed.
	     * (like if they just re-entered...)
	     */
	    update=1;
	    pstatus->type=SP_PSTATUS;
	    pstatus->pnum=i;
	    pstatus->status=pl->p_status;
	    sendClientPacket(pstatus);
	}
	if (me!=pl && flags->flags != htonl(FLAGMASK & pl->p_flags)) {
	    flags->type=SP_FLAGS;
	    flags->pnum=i;
	    flags->flags=htonl(FLAGMASK & pl->p_flags);
	    sendClientPacket(flags);
	}
	if (!hiddenenemy || pl->p_team == me->p_team || 
	      (pl->p_flags & PFSEEN) || (status->tourn==0 && hiddenenemy<2)) {
	    x=pl->p_x;
	    y=pl->p_y;
	} else {
	    /* A hack to make him inviso */
	    x= -100000;
	    y= -100000;
	}
	if (x != ntohl(cpl->x) || y != ntohl(cpl->y) || 
		pl->p_dir != cpl->dir || pl->p_speed != cpl->speed) {
	    /* We update the player if:
	     * 1) haven't updated him for 9 intervals.
	     * 2) he is on the screen
	     * 3) he was on the screen recently.
	     */
	    if (!update && repCount % efticks(9) != 0 && 
		(ntohl(cpl->x) < me->p_x - SCALE*WINSIDE/2 ||
		 ntohl(cpl->x) > me->p_x + SCALE*WINSIDE/2 ||
		 ntohl(cpl->y) > me->p_y + SCALE*WINSIDE/2 ||
		 ntohl(cpl->y) < me->p_y - SCALE*WINSIDE/2) &&
		(y > me->p_y + SCALE*WINSIDE/2 ||
		 x > me->p_x + SCALE*WINSIDE/2 ||
		 x < me->p_x - SCALE*WINSIDE/2 ||
		 y < me->p_y - SCALE*WINSIDE/2)) continue;
	    /* If the guy is cloaked, give information only occasionally,
	     *  and make it slightly inaccurate.
	     * Also, we don't give a direction.  The client has no reason
	     *  to know.
	     */
	    if ((pl->p_flags & PFCLOAK) && 
		  (pl->p_cloakphase == (CLOAK_PHASES - 1)) && 
		  (me != pl) && !mustUpdate[i]) {
		if (repCount % efticks(9) != 0) continue;
		cpl->type=SP_PLAYER;
		cpl->pnum=i;
		cpl->x=htonl(x+(random() % 2000)-1000);
		cpl->y=htonl(y+(random() % 2000)-1000);
		sendClientPacket(cpl);
		continue;
	    }
	    cpl->type=SP_PLAYER;
	    cpl->pnum=i;
	    cpl->x=htonl(x);
	    cpl->y=htonl(y);
	    cpl->speed=pl->p_speed;
	    cpl->dir=pl->p_dir;
	    sendClientPacket(cpl);
	}
    }
}

updateTorps()
{
    register struct torp *torp;
    register int i;
    register struct torp_info_spacket *tpi;
    register struct torp_spacket *tp;

    for (i=0, torp=torps, tpi=clientTorpsInfo, tp=clientTorps; 
	    i<MAXPLAYER*MAXTORP; 
	    i++, torp++, tpi++, tp++) {
	if (torp->t_owner==me->p_no) {
	    if (torp->t_war!=tpi->war ||
		torp->t_status!=tpi->status) {
		tpi->type=SP_TORP_INFO;
		tpi->war=torp->t_war;
		tpi->status=torp->t_status;
		tpi->tnum=htons(i);
		sendClientPacket(tpi);
	    }
	    if (tp->x!=htonl(torp->t_x) ||
		tp->y!=htonl(torp->t_y)) {
		tp->type=SP_TORP;
		tp->x=htonl(torp->t_x);
		tp->y=htonl(torp->t_y);
		tp->dir=torp->t_dir;
		tp->tnum=htons(i);
		sendClientPacket(tp);
	    }
	} else {	/* Someone else's torp... */
	    if (torp->t_y > me->p_y + SCALE*WINSIDE/2 ||
		torp->t_x > me->p_x + SCALE*WINSIDE/2 ||
		torp->t_x < me->p_x - SCALE*WINSIDE/2 ||
		torp->t_y < me->p_y - SCALE*WINSIDE/2 ||
		torp->t_status==TFREE) {
		if (torp->t_status==TFREE && tpi->status==TEXPLODE) {
		    tpi->status=TFREE;
		    continue;
		}
		if (tpi->status!=TFREE) {
		    tpi->status=TFREE;
		    tpi->tnum=htons(i);
		    tpi->type=SP_TORP_INFO;
		    sendClientPacket(tpi);
		}
	    } else {	/* in view */
		if (tp->x!=htonl(torp->t_x) ||
		    tp->y!=htonl(torp->t_y)) {
		    tp->x=htonl(torp->t_x);
		    tp->y=htonl(torp->t_y);
		    tp->dir=torp->t_dir;
		    tp->tnum=htons(i);
		    tp->type=SP_TORP;
		    sendClientPacket(tp);
		}
		if (torp->t_status!=tpi->status ||
		    (torp->t_war & me->p_team)!=tpi->war) {
		    /* Let the client fade away the explosion on its own */
		    tpi->war=torp->t_war & me->p_team;
		    tpi->type=SP_TORP_INFO;
		    tpi->tnum=htons(i);
		    tpi->status=torp->t_status;
		    sendClientPacket(tpi);
		}
	    }
	}
    }
}

updatePlasmas()
{
    register struct plasmatorp *torp;
    register int i;
    register struct plasma_info_spacket *tpi;
    register struct plasma_spacket *tp;

    for (i=0, torp=plasmatorps, tpi=clientPlasmasInfo, tp=clientPlasmas; 
	    i<MAXPLAYER*MAXPLASMA;
	    i++, torp++, tpi++, tp++) {
	if (torp->pt_owner==me->p_no) {
	    if (torp->pt_war!=tpi->war ||
		torp->pt_status!=tpi->status) {
		tpi->type=SP_PLASMA_INFO;
		tpi->war=torp->pt_war;
		tpi->status=torp->pt_status;
		tpi->pnum=htons(i);
		sendClientPacket(tpi);
	    }
	    if (tp->x!=htonl(torp->pt_x) ||
		tp->y!=htonl(torp->pt_y)) {
		tp->type=SP_PLASMA;
		tp->x=htonl(torp->pt_x);
		tp->y=htonl(torp->pt_y);
		tp->pnum=htons(i);
		sendClientPacket(tp);
	    }
	} else {	/* Someone else's torp... */
	    if (torp->pt_y > me->p_y + SCALE*WINSIDE/2 ||
		torp->pt_x > me->p_x + SCALE*WINSIDE/2 ||
		torp->pt_x < me->p_x - SCALE*WINSIDE/2 ||
		torp->pt_y < me->p_y - SCALE*WINSIDE/2 ||
		torp->pt_status==PTFREE) {
		if (torp->pt_status==PTFREE && tpi->status==PTEXPLODE) {
		    tpi->status=PTFREE;
		    continue;
		}
		if (tpi->status!=PTFREE) {
		    tpi->status=PTFREE;
		    tpi->pnum=htons(i);
		    tpi->type=SP_PLASMA_INFO;
		    sendClientPacket(tpi);
		}
	    } else {	/* in view */
		/* Send torp (we assume it moved) */
		tp->x=htonl(torp->pt_x);
		tp->y=htonl(torp->pt_y);
		tp->pnum=htons(i);
		tp->type=SP_PLASMA;
		sendClientPacket(tp);
		if (torp->pt_status!=tpi->status ||
		    (torp->pt_war & me->p_team)!=tpi->war) {
		    tpi->war=torp->pt_war & me->p_team;
		    tpi->type=SP_PLASMA_INFO;
		    tpi->pnum=htons(i);
		    tpi->status=torp->pt_status;
		    sendClientPacket(tpi);
		}
	    }
	}
    }
}

updatePhasers()
{
    register int i;
    register struct phaser_spacket *ph;
    register struct phaser *phase;
    register struct player *pl;

    for (i=0, ph=clientPhasers, phase=phasers, pl=players; 
	 i<MAXPLAYER; i++, ph++, phase++, pl++) {
	if (pl->p_y > me->p_y + SCALE*WINSIDE/2 ||
	    pl->p_x > me->p_x + SCALE*WINSIDE/2 ||
	    pl->p_x < me->p_x - SCALE*WINSIDE/2 ||
	    pl->p_y < me->p_y - SCALE*WINSIDE/2) {
	    if (ph->status!=PHFREE) {
		ph->pnum=i;
		ph->type=SP_PHASER;
		ph->status=PHFREE;
		sendClientPacket(ph);
	    }
	} else {
	    if (phase->ph_status==PHHIT) {
		mustUpdate[phase->ph_target]=1;
	    }
	    if (ph->status!=phase->ph_status ||
		  ph->dir!=phase->ph_dir ||
		  ph->target!=htonl(phase->ph_target)) { 
		ph->pnum=i;
		ph->type=SP_PHASER;
		ph->status=phase->ph_status;
		ph->dir=phase->ph_dir;
		ph->x=htonl(phase->ph_x);
		ph->y=htonl(phase->ph_y);
		ph->target=htonl(phase->ph_target);
		sendClientPacket(ph);
	    }
	}
    }
}

updatePlanets()
{
    register int i;
    register struct planet_spacket *pl;
    register struct planet *plan;
    register struct planet_loc_spacket *pll;

#define PLFLAGMASK (PLREPAIR|PLFUEL|PLAGRI|PLHOME|PLCOUP)

    for (i=0, pl=clientPlanets, plan=planets, pll=clientPlanetLocs; 
	 i<MAXPLANETS;
	 i++, pl++, plan++, pll++) {
	/* Send him info about him not having info if he doesn't but thinks
	 *  he does.
	 * Also send him info on every fifth cycle if the planet needs to
	 *  be redrawn.
	 */
	if (((plan->pl_info & me->p_team)==0) && (pl->info & me->p_team)) {
	    pl->type=SP_PLANET;
	    pl->pnum=i;
	    pl->info=0;
	    sendClientPacket(pl);
	} else 
	if (plan->pl_info & me->p_team &&
	    (pl->info != plan->pl_info ||
	     pl->armies != htonl(plan->pl_armies) ||
	     pl->owner != plan->pl_owner ||
	     pl->flags != htons((short) (plan->pl_flags & PLFLAGMASK)))) {
	    pl->type=SP_PLANET;
	    pl->pnum=i;
	    pl->info=plan->pl_info;
	    pl->flags=htons((short) (plan->pl_flags & PLFLAGMASK));
	    pl->armies=htonl(plan->pl_armies);
	    pl->owner=plan->pl_owner;
	    sendClientPacket(pl);
	} 
	/* Assume that the planet only needs to be updated once... */
	if (pll->x != htonl(plan->pl_x)) {
	    pll->x=htonl(plan->pl_x);
	    pll->y=htonl(plan->pl_y);
	    pll->pnum=i;
	    strcpy(pll->name, plan->pl_name);
	    pll->type=SP_PLANET_LOC;
	    sendClientPacket(pll);
	}
    }
}

updateMessages()
{
    int i;
    struct message *cur;
    struct mesg_spacket msg;
    int updated=0;

    for (i=msgCurrent; i!=(mctl->mc_current+1) % MAXMESSAGE; i=(i+1) % MAXMESSAGE) {
	if (i==MAXMESSAGE) i=0;
	cur= &messages[i];

	if (cur->m_flags & MVALID &&
	    (cur->m_flags & MALL ||
	     (cur->m_flags & MTEAM && cur->m_recpt & me->p_team) ||
	     (cur->m_flags & MINDIV && cur->m_recpt == me->p_no) ||
	     (cur->m_flags & MGOD && cur->m_recpt == me->p_no))) {
	    if (!updated) {
		update_sys_defaults();
		updated=1;
	    }
	    msg.type=SP_MESSAGE;
	    strncpy(msg.mesg, cur->m_data, 80);
	    msg.mesg[79]='\0';
	    msg.m_flags=cur->m_flags;
	    msg.m_recpt=cur->m_recpt;
	    msg.m_from=cur->m_from;
	    sendClientPacket(&msg);
	}
	msgCurrent=(msgCurrent+1) % MAXMESSAGE;
    }
}

sendQueuePacket(pos)
short pos;
{
    struct queue_spacket qPacket;

    qPacket.type=SP_QUEUE;
    qPacket.pos=htons(pos);
    sendClientPacket(&qPacket);
    flushSockBuf();
}

sendClientPacket(packet)
/* Pick a random type for the packet */
struct player_spacket *packet;
{
    int size;

    if (packet->type<1 || packet->type>NUM_SIZES || sizes[packet->type]==0) {
	printf("Attempt to send strange packet %d\n", packet->type);
	return;
    }
    packetsSent[packet->type]++;
    size=sizes[packet->type];
    if (bufptr-buf+size >= BUFSIZE) {
	if (gwrite(sock, buf, bufptr-buf) != bufptr-buf) {
	    clientDead=1;
	}
	bufptr=buf;
    }
    bcopy(packet, bufptr, size);
    bufptr+=size;
}

/* Flush the socket buffer */
flushSockBuf()
{
    if (!clientDead && bufptr != buf) {
	if (gwrite(sock, buf, bufptr-buf) != bufptr-buf) {
	    clientDead=1;
	}
	bufptr=buf;
    }
}

socketPause()
{
    struct timeval timeout;
    int readfds;

    timeout.tv_sec=1;
    timeout.tv_usec=0;
    readfds = 1<<sock;
    select(32,&readfds,0,0,&timeout); 
}

/* Find out if client has any requests */
readFromClient()
{
    struct timeval timeout;
    int readfds;
    char buf[BUFSIZ*2];
    char *bufptr;
    int size;
    int count;
    int temp;

    timeout.tv_sec=0;
    timeout.tv_usec=0;
    readfds = 1<<sock;
    if (clientDead) return(0);
    if (select(32,&readfds,0,0,&timeout) != 0) {
        /* Read info from the xtrek server */
        count=read(sock,buf,BUFSIZ*2);
        if (count<=0) {
	    clientDead=1;
	    return(0);
        }
        bufptr=buf;
        while (bufptr < buf+count) {
	    if (*bufptr < 1 || *bufptr > NUM_PACKETS || handlers[*bufptr].size==0) {
                printf("Unknown packet type: %d, aborting...\n", *bufptr);
                return;
            }
	    size=handlers[*bufptr].size;
	    while (size>count+(buf-bufptr)) {
		/* We wait for up to twenty seconds for rest of packet.
                 * If we don't get it, we assume the client died. 
                 */
		timeout.tv_sec=20;
                timeout.tv_usec=0;
                readfds=1<<sock;
                if (select(32,&readfds,0,0,&timeout) == 0) {
		    logmessage("Died while waiting for packet...");
                    clientDead=1;
                    return(0);
                }

		temp=read(sock,buf+count,size-(count+(buf-bufptr)));
		count+=temp;
		if (temp<=0) {
		    sprintf(buf, "Died in second read(), return=%d", temp);
		    logmessage(buf);
		    clientDead=1;
		    return(0);
		}
	    }
	    /* Check to see if the handler is there and the request is legal.
	     * The code is a little ugly, but it isn't too bad to worry about
	     * yet.
	     */
	    packetsReceived[*bufptr]++;
	    if (handlers[*bufptr].handler != NULL) {
		if (((inputMask==-1 || inputMask==*bufptr) && 
		     (me==NULL || !(me->p_flags & (PFWAR|PFREFITTING)))) || 
		     *bufptr==CP_RESETSTATS || *bufptr==CP_UPDATES ||
		     *bufptr==CP_OPTIONS || *bufptr==CP_RESERVED ||
		     *bufptr==CP_SOCKET || *bufptr==CP_BYE) {
		    if (me && me->p_flags & PFSELFDEST) {
			me->p_flags &= ~PFSELFDEST;
			warning("Self Destruct has been canceled");
		    }
		    (*(handlers[*bufptr].handler))(bufptr);
		}
		/* Otherwise we ignore the request */
	    } else {
		printf("Handler for packet %d not installed...\n", *bufptr);
	    }
            bufptr+=size;
            if (bufptr>buf+BUFSIZ) {
                bcopy(buf+BUFSIZ, buf, BUFSIZ);
                if (count==BUFSIZ*2) {
		    readfds = 1<<sock;
		    if (select(32,&readfds,0,0,&timeout)) {
			temp=read(sock,buf+BUFSIZ,BUFSIZ);
			count=BUFSIZ+temp;
			if (temp<=0) {
			    sprintf(buf, "Died in third read(), return=%d", temp);
			    logmessage(buf);
			    clientDead=1;
			    return(0);
			}
		    } else {
			count=BUFSIZ;
		    }
                } else {
		    count -=BUFSIZ;
		}
                bufptr-=BUFSIZ;
            }
        }
        return(1);
    }
    return(0);
}

handleTorpReq(packet)
struct torp_cpacket *packet;
{
    ntorp(packet->dir, TMOVE);
}

handlePhasReq(packet)
struct phaser_cpacket *packet;
{
    phaser(packet->dir);
}

handleSpeedReq(packet)
struct speed_cpacket *packet;
{
    set_speed(packet->speed);
}

handleDirReq(packet)
struct dir_cpacket *packet;
{
    me->p_flags &= ~(PFPLOCK | PFPLLOCK);
    set_course(packet->dir);
}

handleShieldReq(packet)
struct shield_cpacket *packet;
{
    if (packet->state) {
	shield_up();
    } else {
	shield_down();
    }
}

handleRepairReq(packet)
struct repair_cpacket *packet;
{
    if (packet->state) {
	repair();
    } else {
	me->p_flags &= ~(PFREPAIR);
    }
}

handleOrbitReq(packet)
struct orbit_cpacket *packet;
{
    if (packet->state) {
	orbit();
    } else {
	me->p_flags &= ~PFORBIT;
	if (me->p_flags & PFDOCK) {
	    players[me->p_docked].p_docked--;
	    players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	    me->p_flags &= ~PFDOCK;
	}
    }
}

handlePractrReq(packet)
struct practr_cpacket *packet;
{
    practice_robo();
}

handleBombReq(packet)
struct bomb_cpacket *packet;
{
    if (packet->state) {
	bomb_planet();
    } else {
	me->p_flags &= ~(PFBOMB);
    }
}

handleBeamReq(packet)
struct beam_cpacket *packet;
{
    if (packet->state==1) {
	beam_up();
    } else if (packet->state) {
	beam_down();
    } else {
	me->p_flags &= ~(PFBEAMUP | PFBEAMDOWN);
    }
}

handleCloakReq(packet)
struct cloak_cpacket *packet;
{
    if (packet->state) {
	cloak_on();
    } else {
	cloak_off();
    }
}

handleDetTReq(packet)
struct det_torps_cpacket *packet;
{
    detothers();
}

handleCopilotReq(packet)
struct copilot_cpacket *packet;
{
/*
 * Unsupported...
    if (packet->state) {
	me->p_flags |= PFCOPILOT;
    } else {
	me->p_flags &= ~PFCOPILOT;
    }
 */
}

handleOutfit(packet)
struct outfit_cpacket *packet;
{
    shipPick=packet->ship;
    teamPick=packet->team;
}

sendPickokPacket(state)
int state;
{
    struct pickok_spacket pickPack;

    pickPack.type=SP_PICKOK;
    pickPack.state=state;
    sendClientPacket(&pickPack);
}

handleLoginReq(packet)
struct login_cpacket *packet;
{
    
    strncpy(namePick, packet->name, 16);
    namePick[15]=0;
    strncpy(passPick, packet->password, 16);
    passPick[15]=0;
    /* Is this a name query or a login? */
    if (packet->query) {
	passPick[15]=1;
    }
    strncpy(login, packet->login, 16);
    login[15]=0;
}

sendClientLogin(stats)
struct stats *stats;
{
    struct login_spacket logPacket;

    logPacket.type=SP_LOGIN;
    if (stats==NULL) {
	logPacket.accept=0;
    } else {
	logPacket.accept=1;
	logPacket.flags=htonl(stats->st_flags);
	bcopy(stats->st_keymap, logPacket.keymap, 96);
    }
    sendClientPacket(&logPacket);
}

handlePlasmaReq(packet)
struct plasma_cpacket *packet;
{
    nplasmatorp(packet->dir, PTMOVE);
}

handleWarReq(packet)
struct war_cpacket *packet;
{
    declare_war(packet->newmask);
}

handlePlanlockReq(packet)
struct planlock_cpacket *packet;
{
    lock_planet(packet->pnum);
}

handlePlaylockReq(packet)
struct playlock_cpacket *packet;
{
    lock_player(packet->pnum);
}

handleDetMReq(packet)
struct det_mytorp_cpacket *packet;
{
    struct torp *atorp;
    int t;

    t=ntohs(packet->tnum);
    if (t<0 || t>=MAXPLAYER*MAXTORP) return;
    atorp= &torps[t];
    
    if (atorp->t_owner != me->p_no) return;
    if (atorp->t_status == TMOVE || atorp->t_status == TSTRAIGHT) {
	atorp->t_status = TOFF;
    }
}

handleTractorReq(packet)
struct tractor_cpacket *packet;
{
    int target;
    struct player *player;

    if (weaponsallowed[WP_TRACTOR]==0) {
	warning("Tractor beams haven't been invented yet.");
	return;
    }
    target=packet->pnum;
    if (packet->state==0) {
	me->p_flags &= ~(PFTRACT | PFPRESS);
	return;
    }
    if (me->p_flags & PFCLOAK) {
	warning("Weapons's Officer:  Cannot tractor while cloaked, sir!");
	return;
    }
    if (target<0 || target>=MAXPLAYER || target==me->p_no) return;
    player= &players[target];
    if (player->p_flags & PFCLOAK) return;
    if (hypot((double) me->p_x-player->p_x,
	    (double) me->p_y-player->p_y) < 
	    ((double) TRACTDIST) * me->p_ship.s_tractrng) {
	if (player->p_flags & PFDOCK) {
	    players[player->p_docked].p_port[player->p_port[0]] = VACANT;
	    players[player->p_docked].p_docked--;
	}
	if (me->p_flags & PFDOCK) {
	    players[me->p_docked].p_docked--;
	    players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	}
	player->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_tractor = target;
	me->p_flags |= PFTRACT;
    } else {
	warning("Weapon's Officer:  Vessel is out of range of our tractor beam.");
    }
}

handleRepressReq(packet)
struct repress_cpacket *packet;
{
    int target;
    struct player *player;

    if (weaponsallowed[WP_TRACTOR]==0) {
	warning("Tractor beams haven't been invented yet.");
	return;
    }
    target=packet->pnum;
    if (packet->state==0) {
        me->p_flags &= ~(PFTRACT | PFPRESS);
        return;
    }
    if (me->p_flags & PFCLOAK) {
	warning("Weapons's Officer:  Cannot tractor while cloaked, sir!");
	return;
    }
    if (target<0 || target>=MAXPLAYER || target==me->p_no) return;
    player= &players[target];
    if (player->p_flags & PFCLOAK) return;
    if (hypot((double) me->p_x-player->p_x,
	    (double) me->p_y-player->p_y) < 
	    ((double) TRACTDIST) * me->p_ship.s_tractrng) {
	if (player->p_flags & PFDOCK) {
	    players[player->p_docked].p_port[player->p_port[0]] = VACANT;
	    players[player->p_docked].p_docked--;
	}
	if (me->p_flags & PFDOCK) {
	    players[me->p_docked].p_docked--;
	    players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	}
	player->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_tractor = target;
	me->p_flags |= (PFTRACT | PFPRESS);
    } else {
	warning("Weapon's Officer:  Vessel is out of range of our tractor beam.");
    }
}

sendMotdLine(line)
char *line;
{
    struct motd_spacket motdPacket;

    motdPacket.type=SP_MOTD;
    strncpy(motdPacket.line, line, 80);
    motdPacket.line[79]='\0';
    sendClientPacket(&motdPacket);
}

handleCoupReq(packet)
struct coup_cpacket *packet;
{
    coup();
}

handleRefitReq(packet)
struct refit_cpacket *packet;
{
    do_refit(packet->ship);
}

handleMessageReq(packet)
struct mesg_cpacket *packet;
{
    char addrbuf[9];
    static long lasttime=0, time();
    static int balance=0;	/* make sure he doesn't get carried away */
    long thistime;

    /* Some random code to make sure the player doesn't get carried away 
     *  about the number of messages he sends.  After all, he could try to 
     *  jam peoples communications if we let him.
     */
    thistime=time(NULL);
    if (lasttime!=0) {
	balance=balance-(thistime-lasttime);
	if (balance<0) balance=0;
    }
    lasttime=thistime;
    if (balance>=15) {
	warning("Be quiet");
	balance+=3;
	if (balance > time(NULL) + 60) {
	    balance=time(NULL)+60;
	}
	return;
    }
    balance+=3;
    packet->mesg[69]='\0';
    sprintf(addrbuf, " %c%c->", teamlet[me->p_team], shipnos[me->p_no]);
    switch(packet->group) {
    case MALL:
	sprintf(addrbuf+5, "ALL");
	break;
    case MGOD:
	sprintf(addrbuf+5, "GOD");
	break;
    case MTEAM:
	if (packet->indiv != FED && packet->indiv != ROM &&
	    packet->indiv != KLI && packet->indiv != ORI) return;
	sprintf(addrbuf+5, teamshort[packet->indiv]);
	break;
    case MINDIV:
	if (packet->indiv < 0 || packet->indiv >= MAXPLAYER) return;
	if (players[packet->indiv].p_status == PFREE) return;
	sprintf(addrbuf+5, "%c%c ", teamlet[players[packet->indiv].p_team],
	    shipnos[packet->indiv]);
	break;
    default:
	return;
    }
    pmessage2(packet->mesg, packet->indiv, packet->group, addrbuf, me->p_no);
}

handleQuitReq(packet)
struct quit_cpacket *packet;
{
    me->p_flags |= PFSELFDEST;
    selfdest = me->p_updates + 100;
    /* 60 seconds to destruct a starbase */
    if (me->p_ship.s_type==STARBASE) selfdest = me->p_updates + 600;
    warning("Self destruct initiated");
}

sendMaskPacket(mask)
int mask;
{
    struct mask_spacket maskPacket;

    maskPacket.type=SP_MASK;
    maskPacket.mask=mask;
    sendClientPacket(&maskPacket);
}

handleOptionsPacket(packet)
struct options_cpacket *packet;
{
    bcopy(packet->keymap, mystats->st_keymap, 96);
    mystats->st_flags = ntohl(packet->flags);
    keeppeace=(mystats->st_flags / ST_KEEPPEACE) & 1;
}

handleSocketReq(packet)
struct socket_cpacket *packet;
{
    nextSocket=ntohl(packet->socket);
    userVersion=packet->version;
}

handleByeReq(packet)
struct bye_cpacket *packet;
{
    noressurect=1;
}

checkVersion()
{
    struct badversion_spacket packet;

    if (userVersion != SOCKVERSION) {
	packet.type=SP_BADVERSION;
	packet.why=0;
	sendClientPacket(&packet);
	flushSockBuf();
	return(0);
    }
    return(1);
}

logEntry()
{
    FILE *logfile;
    int curtime;

    logfile=fopen(LOGFILENAME, "a");
    if (!logfile) return;
    curtime=time(NULL);
    fprintf(logfile, "Joining: %s (%s@%s) (%c), %s", me->p_name,
	me->p_login, me->p_monitor,			/* ATM */
	shipnos[me->p_no], ctime(&curtime));
    fclose(logfile);
}

gwrite(fd, buf, bytes)
int fd;
char *buf;
register int bytes;
{
    long orig = bytes;
    register long n;
    char tempbuf[80];
    struct timeval to;

    while (bytes) {
	n = write(fd, buf, bytes);
	if (n < 0) {
	    if (errno==ENOBUFS) {
		/* The man pages don't mention this as a possibility.
		 * Yet, it happens.  I guess I just wait for 1/10 sec, and
		 *  continue?
		 */
		/* I would use usleep() to do this, but this system ain't
		 *  got it...
		 */
		to.tv_sec=0;
		to.tv_usec=100000;
		select(0,NULL,NULL,NULL,&to);
		continue;
	    }
	    sprintf(tempbuf, "Died in gwrite, n=%d, errno=%d", n, errno);
	    logmessage(tempbuf);
	    return(-1);
	}
	bytes -= n;
	buf += n;
    }
    return(orig);
}

handleDockingReq(packet)
struct dockperm_cpacket *packet;
{
    int i;

    if (me->p_ship.s_type == STARBASE) {
	me->p_docked = 0;
	for (i=0; i<NUMPORTS; i++) {
	    if (me->p_port[i] != VACANT)
		players[me->p_port[i]].p_flags &= ~PFDOCK;
	    me->p_port[i] = VACANT;  
	}
	if (packet->state) 
	    me->p_flags |= PFDOCKOK;
	else
	    me->p_flags &= ~PFDOCKOK;
    }
}

handleReset(packet)
struct resetstats_cpacket *packet;
{
    if (packet->verify != 'Y') return;

    /* Gee, they seem to want to reset their stats!  Here goes... 
     */
    mystats->st_maxkills=0.0;
    mystats->st_kills=0;
    mystats->st_losses=0;
    mystats->st_armsbomb=0;
    mystats->st_planets=0;
    mystats->st_ticks=0;
    mystats->st_tkills=0;
    mystats->st_tlosses=0;
    mystats->st_tarmsbomb=0;
    mystats->st_tplanets=0;
    mystats->st_tticks=1;
    mystats->st_rank=0;
    mystats->st_sbkills=0;
    mystats->st_sblosses=0;
    mystats->st_sbticks=0;
    mystats->st_sbmaxkills=0.0;
}

handleUpdatesReq(packet)
struct updates_cpacket *packet;
{
    struct itimerval	udt;
    struct timeval timeout;
    int readfds;

    timerDelay=ntohl(packet->usecs);
    if (timerDelay < 100000) timerDelay=100000;
    if (timerDelay >= 1000000) timerDelay=999999;

    udt.it_interval.tv_sec = 0;
    udt.it_interval.tv_usec = timerDelay;
    udt.it_value.tv_sec = 0;
    udt.it_value.tv_usec = timerDelay;
    setitimer(ITIMER_REAL, &udt, 0);
}

logmessage(string)
char *string;
{
    FILE *fp;

    fp=fopen(LOGFILENAME, "a");
    if (fp) {
	fprintf(fp, "%s\n", string);
	fclose(fp);
    }
}

handleReserved(packet)
struct reserved_cpacket *packet;
{
    struct reserved_cpacket mycp;
    struct reserved_spacket mysp;

    if (testtime==1) return;
    if (bcmp(packet->data, testdata, 16) != 0) {
	testtime=1;
	return;
    }
    bcopy(testdata, mysp.data, 16);
    encryptReservedPacket(&mysp, &mycp, serverName, me->p_no);
    if (bcmp(packet->resp, mycp.resp, 16) != 0) {
	fprintf(stderr, "User verified incorrectly.\n");
	testtime=1;
	return;
    }
    testtime=0;
}
